//
// Copyright (c) 2018 INRIA
//

#ifndef __pinocchio_dynamics_hpp__
#define __pinocchio_dynamics_hpp__

#include "pinocchio/macros.hpp"

PINOCCHIO_PRAGMA_DEPRECATED_HEADER(pinocchio/algorithm/dynamics.hpp,pinocchio/algorithm/contact-dynamics.hpp)

#include "pinocchio/algorithm/contact-dynamics.hpp"

#endif // ifndef __pinocchio_dynamics_hpp__
