//
// Copyright (c) 2019 INRIA
//

#ifndef __pinocchio_python_utils_dependencies_hpp__
#define __pinocchio_python_utils_dependencies_hpp__

#include "pinocchio/bindings/python/fwd.hpp"

namespace pinocchio
{
  namespace python
  {
    
    void exposeDependencies();
    
  } // namespace python
} // namespace pinocchio

#endif // ifndef __pinocchio_python_utils_dependencies_hpp__
